/*
 *   CENTER FOR THEORY AND SIMULATION IN SCIENCE AND ENGINEERING
 *                      CORNELL UNIVERSITY
 *
 *      Portions of this software may fall under the following
 *      copyrights: 
 *
 *      Copyright (c) 1983 Regents of the University of California.
 *      All rights reserved.  The Berkeley software License Agreement
 *      specifies the terms and conditions for redistribution.
 *
 *      The copyright (c) of the EGPUP User Process, Paul Kirton and
 *      ISI, August 23, 1984
 *
 */

/*
 *      The following acknowledgments and thanks apply:
 *
 *      Mike Petry (petry@trantor.umd.edu), University of Maryland, for
 *      development of the HELLO routines used in this program.
 *
 *      Craig Partridge (craig@loki.bbn.com), BBN Laboratories, for
 *      thorough linting, profiling, and performance enhancements to
 *      this program.
 *
 *      Charles Hedrick (hedrick@topaz.rutgers.edu),
 *      Chris Maio (chris@columbia.edu), Sergio Heker (heker@jvnca.csc.org),
 *      and any other BETA testers who have helped in finding bugs.
 *
 *	Mark Fedor (now fedor@nisc.nyser.net) for the development and
 *	maintenance up to release 1.3.1.  Also for his continuing
 *	advice.
 *
 *	Thomas Narten <narten@purdue.edu> for his analysis and
 *	modifications to the EGP and other helpful improvements.
 *
 *
 *
 *  GATED - based on Kirton's EGP, UC Berkeley's routing daemon (routed),
 *          and DCN's HELLO routing Protocol.  HELLO output routines
 *          developed and originally written at the University of
 *          Maryland by Mike Petry.
 * 
 *  December 5, 1986 - Mark Fedor
 *
 */

#ifndef	lint
static char *rcsid = "$Header: egp.c,v 1.3.1.18 88/02/10 19:59:37 jch Locked $";
#endif	not lint

#include        "include.h"

/*
 * The major EGP state changes and initializations occur when the
 * gateway first comes up, init_egp() (in init.c), when a neighbor goes into
 * state NEIGHBOR, egpstngh(), and when a neighbor goes into state UNACQUIRED,
 * egpstunacq().
 */

/*
 * egpjob(): periodic EGP functions and packet sending.
 *
 * egpjob() is called periodically by timeout() after each timer interrupt. It
 * checks each neighbor's state table in turn and decides what action to take
 * based on the state and time elapsed. It initiates the sending of 
 * Acquisition Request, Hello and Poll commands and the retransmission of 
 * Acquisition Request, Poll and Cease commands. Appropriate functions are 
 * called to format each of these  messages.  In the NEIGHBOR state it 
 * determines the the current reachability of the peer. After 480 seconds have
 * elapsed it calls egpchkcmd() to check whether peer commands are being 
 * received at an excess rate.
 *
 * At most maxacq neighbors will be acquired. The number of new acq. requests
 * sent this interrupt does not exceed the number of neighbors yet to be
 * acquired. Any number of retransmitted acq. requests may be sent to 
 * non-responding neighbors.
 */

egpjob()
{
  register egpng *ngp;
  register int   shiftreg;
  static long    chkcmdtime = 0;
  int   n_new_acqsnt = 0,         /* # new acq. requests sent this interrupt */
                       i,
                       nresponses;

  /* commence periodic processing for each egp neighbor */

  for (ngp = egpngh; ngp != NULL; ngp = ngp->ng_next) {
    switch(ngp->ng_state) {
      case UNACQUIRED:
        if (terminate)
          break;
        if (n_new_acqsnt >= (maxacq - n_acquired))
          break;
        if (gatedtime < ngp->ng_htime)
          break;
        ngp->ng_state = ACQUIRE_SENT;
        ngp->ng_htime = gatedtime + ngp->ng_hint;
        egpsacq(ngp->ng_addr, ngp->ng_myaddr,NAREQ, ACTIVE, ngp->ng_sid);
        n_new_acqsnt++;
        break;
      case ACQUIRE_SENT:
        if (gatedtime < ngp->ng_htime)
          break;
        if (++ngp->ng_retry >= NACQ_RETRY) {
          ngp->ng_hint = LONGACQINT;
          egpstime();
          ngp->ng_retry = 0;
        }
        ngp->ng_htime = gatedtime + ngp->ng_hint;
        egpsacq(ngp->ng_addr, ngp->ng_myaddr, NAREQ, ACTIVE, ngp->ng_sid);
        break;
      case NEIGHBOR:
        /* determine reachability */
        if (gatedtime >= ngp->ng_htime) {
          shiftreg = ngp->ng_responses;
          nresponses = 0;
          for (i = 0; i < NCOMMANDS; i++) {     /* count responses */
             if (shiftreg & 1)
               nresponses++;
             shiftreg >>= 1;
          }
          if (ngp->ng_reach & NG_DOWN) {
            if (nresponses > NRESPONSES) {      /* down -> up */
	     syslog(LOG_WARNING, "EGP neighbor %s state change DOWN -> UP, %d/%d responses received",
		    inet_ntoa(ngp->ng_addr), nresponses, NCOMMANDS);
              ngp->ng_reach &= ~NG_DOWN;
              ngp->ng_retry = 0;
              TRACE_EXT("egpjob: %s reachable %s\n",
                            inet_ntoa(ngp->ng_addr), strtime);
            }
            else                            /* down -> down */
              if (++ngp->ng_retry >= NUNREACH) {
                egpcease(ngp, UNSPEC);
                break;
              }
          }
          else
            if (nresponses < NRESPONSES) {  /* up -> down */
	     syslog(LOG_WARNING, "EGP neighbor %s state change UP -> DOWN, %d/%d responses received",
		    inet_ntoa(ngp->ng_addr), nresponses, NCOMMANDS);
              ngp->ng_reach |= NG_DOWN;
              TRACE_EXT("egpjob: %s unreachable %s\n",
                               inet_ntoa(ngp->ng_addr), strtime);
              rt_unreach(ngp->ng_addr); /* delete routes for down gateway */
              if (maxacq < nneigh ) {
                egpcease(ngp, UNSPEC);
                break;
              }
              else {
                ngp->ng_retry = 1;
              }
            }       
        }
        /*
         * Send NR poll
         */
        if (gatedtime >= ngp->ng_stime) {       /* time for poll */
          if (ngp->ng_reach == BOTH_UP) {
            if (ngp->ng_snpoll < NPOLL) {
              if (ngp->ng_snpoll++ == 0) {       /* not repoll */
                ngp->ng_sid++;
                ngp->ng_runsol = 0;
                ngp->ng_noupdate++;
              }
              /* repoll interval is hello int */
              ngp->ng_stime = gatedtime + ngp->ng_hint;
              ngp->ng_responses <<= 1;
              if (ngp->ng_noupdate <= MAXNOUPDATE)
                egpspoll(ngp);
              else {
                TRACE_EXT("egpjob: no valid NR update");
                TRACE_EXT(" for %d successive new poll", MAXNOUPDATE);
                TRACE_EXT(" id's\n");
                egpcease(ngp, UNSPEC);
              }
              break;
            }
            else {          /* no more repolls */
              ngp->ng_stime += ngp->ng_spint - ngp->ng_hint;
              ngp->ng_snpoll = 0;
            }
          }
          else
            ngp->ng_snpoll = 0; /* reach. maybe changed when poll outstanding */
        }
        /*
         * Send hello
         */
        if (gatedtime >= ngp->ng_htime) {
          ngp->ng_htime = gatedtime + ngp->ng_hint;
          ngp->ng_responses <<= 1;
          egpshello(ngp, HELLO, ngp->ng_sid);
        }
        break;
      case CEASE_SENT:
        if (gatedtime < ngp->ng_htime)
          break;
        if (++ngp->ng_retry > NCEASE_RETRY) {
          egpstunacq(ngp);
          break;
        }
        ngp->ng_htime = gatedtime + ngp->ng_hint;
        egpsacq(ngp->ng_addr,ngp->ng_myaddr,NACEASE,ngp->ng_status,ngp->ng_sid);
        break;
      default:
        break;
    }
  }
  if (gatedtime >= chkcmdtime) {
    if (chkcmdtime)                 /* ignore initialization */
      egpchkcmd((u_long)gatedtime - chkcmdtime);
    chkcmdtime = gatedtime + CHKCMDTIME;
  }
  return;
}

/*
 * egpchkcmd() check command reception rate for neighbors and if too high
 * cease and go to state BAD so no further peering.
 */

egpchkcmd( extratime)
        u_long  extratime;
{
  reg egpng *ngp;
  u_long  cmdtime;                /* time since last check # commands recvd */
  u_long  nmaxcmd;                /* max # commands allowed */

  cmdtime = extratime + CHKCMDTIME;      /* compute max. allowed cmds */
  nmaxcmd = (NMAXCMD * cmdtime) / CHKCMDTIME;

  /*
   * For each neighbor check if too many cmds recvd
   */
  for (ngp = egpngh; ngp != NULL; ngp = ngp->ng_next) {
    if (ngp->ng_rcmd > nmaxcmd) {
      ngp->ng_flags |= NG_BAD;
      TRACE_EXT("egpjob: recv too many commands (%d) in ", ngp->ng_rcmd);
      TRACE_EXT("%d sec from %s\n", cmdtime, inet_ntoa(ngp->ng_addr));
      egpcease(ngp, PARAMPROB);
    }
    ngp->ng_rcmd = 0;
  }
}

/*
 * egpin() handles received EGP packets. 
 *
 * It first checks for fragmented packets, valid EGP checksum, version, and
 * length. If any of these are in error the packet is discarded. If the source
 * and destination addresses match those expected in the state tables the
 * appropriate function is called according to the EGP message type. Otherwise
 * the appropriate Refuse, Cease-ack, Error or Cease message is returned.
 * Received Error messages are merely logged with no other action taken.
 */

egpin(pkt)
        char *pkt;
{
  register     struct  ip      *ip = (struct ip *)pkt;
  register     egpng           *ngp;
  register     egpkt           *egp;
  int          iphlen, egplen;

  if (ip->ip_off & ~IP_DF) {
    TRACE_EXT("egpin: recv fragmanted pkt from %s\n", inet_ntoa(ip->ip_src));
    return;
  }
  iphlen = ip->ip_hl;
  iphlen <<= 2;
  egp = (egpkt *) (pkt + iphlen);
  egplen = ip->ip_len;

  if (egp_cksum((char *)egp, egplen) != 0) {
    TRACE_EXT("egpin: bad checksum from %s\n", inet_ntoa(ip->ip_src));
    syslog(LOG_WARNING, "Bad EGP checksum from %s", inet_ntoa(ip->ip_src));
    return;
  }
  if (egp->egp_ver != EGPVER) {
    TRACE_EXT("egpin: bad version %d from %s\n", egp->egp_ver,
                                        inet_ntoa(ip->ip_src));
    return;
  }
  if (egplen < EGPLEN) {
    syslog(LOG_WARNING, "Bad pkt length %d from %s", egplen,
	   inet_ntoa(ip->ip_src));
    TRACE_EXT("egpin: bad pkt length %d from %s\n", egplen,
                                        inet_ntoa( ip->ip_src));
    return;
  }
  egprid_h = ntohs(egp->egp_id);  /* save sequence number in host byte order */

  /*
   * check if in legitimate neighbor list
   */
  for (ngp=egpngh; ngp != NULL; ngp = ngp->ng_next) {    /* all legit neigh */
    if (ngp->ng_state == EMPTY)
      continue;
    if (ngp->ng_addr.s_addr == ip->ip_src.s_addr) {      /* legit neigh */
      if (ngp->ng_myaddr.s_addr != ip->ip_dst.s_addr &&
          egp->egp_type != EGPERR) {
        egpserr(ip->ip_src,ip->ip_dst, egp, egplen, EUNSPEC, (egpng *)0);
        return;
      }
      if (ngp->ng_state == BAD)
        return;
      switch (egp->egp_type) {
        case EGPACQ:
          egpacq(ngp, egp, egplen);
          return;
        case EGPHELLO:
          egphello(ngp, egp);
          return;
        case EGPNR:
          egpnr(ngp, egp, egplen);
          return;
        case EGPPOLL:
          egppoll(ngp, egp);
          return;
        case EGPERR:
          if (ngp->ng_state == NEIGHBOR) {
	     syslog(LOG_WARNING, "Error packet from EGP neighbor %s",
		    inet_ntoa(ngp->ng_addr));
            if (egp->egp_status == DOWN) {
		syslog(LOG_WARNING, "EGP neighbor %s thinks we are down",
		    inet_ntoa(ngp->ng_addr));
              ngp->ng_reach |= ME_DOWN;
            } else {
              ngp->ng_reach &= ~ME_DOWN;
	    }
            ngp->ng_responses |= 1;
          }
          break;
        default:
          egpserr(ip->ip_src,ip->ip_dst, egp, egplen, EBADHEAD, (egpng *)0);
          return;
      }                               /* end switch type */
      break;
    }                                 /* end if legit */
  }                                   /* end for all legit neigh */
  switch (egp->egp_type) {            /* May be legit. neighbor */
    case EGPERR:
      if (tracing & TR_EXT) {
        if (tracing & TR_PKT)
          printf("egpin: recv above error packet\n");
        else
          traceegp("egpin: recv error pkt: ", ip->ip_src,ip->ip_dst,egp,egplen);
      }
      break;
    case EGPACQ:        /* Not legit. neighbor */
      switch (egp->egp_code) {
        case NAREQ:
          egpsacq(ip->ip_src, ip->ip_dst, NAREFUS, ADMINPROHIB, egprid_h);
          break;
        case NACEASE:
          egpsacq(ip->ip_src, ip->ip_dst, NACACK, egp->egp_status, egprid_h);
          break;
        case NACACK:
          break;
        case NACONF:
        default:
          egpsacq(ip->ip_src, ip->ip_dst, NACEASE, PROTOVIOL, egprid_h);
          break;
      }
      break;
    case EGPNR:
    case EGPHELLO:
    case EGPPOLL:
    default:
      egpsacq(ip->ip_src, ip->ip_dst, NACEASE, PROTOVIOL, egprid_h);
      break;
  }                                  /* end switch type */
  return;
}

/*
 * egpacq() handles received Neighbor Acquisition messages: Request, Confirm,
 * Refuse, Cease and Cease-ack. 
 *
 * It calls egpsetint() to validate and set Hello and Poll intervals, 
 * egpstngh() to initialize state tables when the state changes to ACQUIRED,
 * and egpstunacq() when the state changes to UNACQUIRED. The last two
 * functions call egpstime() to recompute the interrupt timer interval.
 */

egpacq(ngp, egp, egplen)
	register     egpng   *ngp;
	register     egpkt   *egp;
        int     egplen;
{
  u_short error = 0;

  switch (egp->egp_code) {
    case NAREQ:                     /* Neighbor acquisition request */
      ngp->ng_rcmd++;
      if (egplen < sizeof(struct egpacq)) {
        error = 1;
        break;
      }
      if (ngp->ng_state == CEASE_SENT) 
        egpsacq(ngp->ng_addr, ngp->ng_myaddr, NACEASE, ngp->ng_status,egprid_h);
      else
        if (terminate) 
          egpsacq(ngp->ng_addr, ngp->ng_myaddr, NAREFUS, GODOWN, egprid_h);
      else
        if (ngp->ng_state != NEIGHBOR && n_acquired == maxacq)
          egpsacq(ngp->ng_addr, ngp->ng_myaddr, NAREFUS, NORESOURCE, egprid_h);
      else
        if (error = egpsetint(ngp, egp))       /* intervals too big */
          egpsacq(ngp->ng_addr, ngp->ng_myaddr, NAREFUS, PARAMPROB, egprid_h);
      else {
        egpstngh(ngp, egp);
        ngp->ng_rid = egprid_h;
        egpsacq(ngp->ng_addr, ngp->ng_myaddr, NACONF, ACTIVE, egprid_h);
      }
      break;
    case NACONF:                    /* Neighbor acq. confirm */
      if (ngp->ng_state == ACQUIRE_SENT && egprid_h == ngp->ng_sid &&
          egplen >= sizeof(struct egpacq)) {
        if (error = egpsetint(ngp, egp)) {      /* intervals too big */
          ngp->ng_flags |= NG_BAD;              /* delay acq retry */
          egpcease(ngp, PARAMPROB);
        }
        else {
          egpstngh(ngp, egp);
          ngp->ng_rid = egprid_h;
        }
      }
      else {
        if (ngp->ng_state == UNACQUIRED)
          egpsacq(ngp->ng_addr, ngp->ng_myaddr, NACEASE, PROTOVIOL, egprid_h);
        error = 1;
      }
      break;
    case NAREFUS:                   /* Neighbor acq. refuse */
      if (ngp->ng_state == ACQUIRE_SENT && egprid_h == ngp->ng_sid) {
        syslog(LOG_WARNING, "Neighbor aquistion refuse from EGP neighbor %s",
	       inet_ntoa(ngp->ng_addr));
        egpstunacq(ngp);
      } else
        error = 1;
      break;
    case NACEASE:                   /* Neighbor acq. cease */
      ngp->ng_rcmd++;
      if (ngp->ng_state == NEIGHBOR) {
	  syslog(LOG_WARNING, "Cease command from EGP neighbor %s\n",
	   inet_ntoa(ngp->ng_addr));
        TRACE_EXT("egpacq: cease from %s, reason %d, %s\n",
                       inet_ntoa(ngp->ng_addr), egp->egp_status, strtime);
      }
      egpstunacq(ngp);
      egpsacq(ngp->ng_addr, ngp->ng_myaddr, NACACK, egp->egp_status, egprid_h);
      return;
    case NACACK:                    /* Neighbor acq. cease ack */
      if (ngp->ng_state == CEASE_SENT && egprid_h == ngp->ng_sid)
        egpstunacq(ngp);
      else    
        /*
         * in unacq. state cease may be sent as response to invalid
         * packets so may get valid ceaseack
         */
        if (ngp->ng_state != UNACQUIRED)
          error = 1;
      break;
    default:
      error = 1;
      break;
  }
  if (error && tracing & TR_EXT) {
    if (!(tracing & TR_PKT))
      traceegp("egpacq: recv bad pkt: ",
                        ngp->ng_addr, ngp->ng_myaddr, egp, egplen);
    TRACE_EXT("egpacq: last recvd pkt bad, recvd in state %d\n", ngp->ng_state);
  }
  return;
}

/*
 * egpsetint() sets EGP hello and poll intervals and times.
 * Returns 1 if either poll or hello intervals too big, 0 otherwise.
 */

egpsetint(ngp, egppkt)
        register     egpng   *ngp;
        register     struct  egppkt  *egppkt;
{
  register     struct  egpacq  *egpa = (struct egpacq *)egppkt;
  u_short helloint, pollint, ratio;

  /*
   * check parameters within bounds
   */
  helloint = ntohs(egpa->ea_hint);
  pollint  = ntohs(egpa->ea_pint);
  if (helloint > MAXHELLOINT || pollint > MAXPOLLINT) {
    TRACE_EXT("Hello interval = %d or poll interval = %d too big",
                        helloint, pollint);
    TRACE_EXT(" from %s, code %d,\n%s", inet_ntoa(ngp->ng_addr),
                        egpa->ea_pkt.egp_code, strtime);
    syslog(LOG_WARNING, "EGP neighbor %s specified intervals too large for us",
	   inet_ntoa(ngp->ng_addr));
    return(1);
  }
  if ((helloint != MINHELLOINT) || (pollint != MINPOLLINT))
      syslog(LOG_WARNING,
	     "EGP neighbor %s specified hello/poll intervals %d/%d, we specified %d/%d",
	     inet_ntoa(ngp->ng_addr), helloint, pollint, MINHELLOINT,
	     MINPOLLINT);
  if (helloint < MINHELLOINT)
      helloint = MINHELLOINT;
  if (pollint < MINPOLLINT)
      pollint = MINPOLLINT;
  ratio = (pollint - 1) / helloint + 1;   /* keep ratio pollint:helloint */
  helloint += HELLOMARGIN;
  pollint = ratio * helloint;
  syslog(LOG_WARNING, "Using actual intervals %d/%d with EGP neighbor %s",
	 helloint, pollint, inet_ntoa(ngp->ng_addr));
  ngp->ng_hint = helloint;
  ngp->ng_htime = gatedtime + helloint;
  ngp->ng_spint = helloint * ratio;
  ngp->ng_stime = gatedtime + helloint;
  if (helloint < egpsleep)
      egpsleep = helloint;
#ifdef debug
  printf("egpsetint: intervals hello %d, poll %d, egpsleep %d\n",
	 ngp->ng_hint, ngp->ng_spint, egpsleep);
#endif  
  return(0);
}

/*
 * egpstngh() go into neighbor state, initialize most variables.
 */

/* ARGSUSED */
egpstngh(ngp, egp)
	register     egpng   *ngp;
	register     egpkt   *egp;
{
#ifdef lint
  egp = egp;    /* make lint happy */
#endif lint

  if (ngp->ng_state != NEIGHBOR) {
    n_acquired++;
    TRACE_EXT("egpstngh: acquired %s %s\n", inet_ntoa(ngp->ng_addr), strtime);
    syslog(LOG_WARNING,"Acquired EGP neighbor %s", inet_ntoa(ngp->ng_addr));
  }
  ngp->ng_state = NEIGHBOR;
  ngp->ng_reach = ME_DOWN;        /* so don't send initial poll until
                                   * after peer reports me up, but I will
                                   * respond to peer polls immediately 
                                   * as a poll implies he thinks I am up
                                   */
  ngp->ng_responses = ~0;         /* assume neighbor up when acquired */
  ngp->ng_flags = 0;
  ngp->ng_status = 0;
  ngp->ng_retry = 0;
  ngp->ng_snpoll = 0;
  ngp->ng_runsol = 0;
  ngp->ng_noupdate = 0;
  ngp->ng_rpint = MINPOLLINT;
  ngp->ng_rnpoll = 0;
  ngp->ng_rtime = gatedtime;

  if (n_acquired == maxacq) 
    egp_maxacq();

  egpstime();
}

/*
 * egp_maxacq() ceases any neighbors with outstanding acquisition requests
 * once the maximum number of neighbors have been acquired. It also resets
 * all acquisition intervals to the short value.
 */

egp_maxacq()
{
  struct egpngh *ngp;

  for (ngp = egpngh; ngp != NULL; ngp = ngp->ng_next) {
    if (ngp->ng_state == ACQUIRE_SENT)
      egpcease(ngp, NORESOURCE);
    if (ngp->ng_state == UNACQUIRED)
      ngp->ng_hint = MINHELLOINT + HELLOMARGIN;
  }
  return;
}

/*
 * egpstime() sets egpsleep, the time between periodic sending of EGP packets;
 * maxpollint - the maximum poll interval for all acquired neighbors;
 * and rt_maxage - the maximum life time of routes. egpsleep is set to the
 * minimum interval for all potential peers or 60 s, whichever is less.
 * rt_maxage is set to the maximum of RT_MINAGE and RT_NPOLLAGE times the
 * maximum poll interval.
 * Called by egpstunacq(), egpstngh() and egpjob()
 */

egpstime()
{
  register     egpng   *ngp;

  maxpollint = MINPOLLINT;

  for (ngp = egpngh; ngp != NULL; ngp = ngp->ng_next) {
    switch (ngp->ng_state) {
      case NEIGHBOR:
        if (ngp->ng_spint > maxpollint)
          maxpollint = ngp->ng_spint;
      case UNACQUIRED:
      case ACQUIRE_SENT:
      case CEASE_SENT:
        if (ngp->ng_hint < egpsleep)
          egpsleep = ngp->ng_hint;
        break;
      default:
        break;
    }
  }
  if (egpsleep < MINHELLOINT) {
    TRACE_EXT("egpstime: bad sleep time %d\n", egpsleep);
    egpsleep = MINHELLOINT + HELLOMARGIN;
  }
  if (n_acquired > 0) /* only change rt_maxage if have an acquired neighbor */
    rt_maxage = (maxpollint * RT_NPOLLAGE > RT_MINAGE) ? 
                           (maxpollint * RT_NPOLLAGE) : RT_MINAGE;
  return;
}

/*
 * egpstunacq() sets state as UNACQUIRED.
 *
 * External variables:
 * terminate
 */

egpstunacq(ngp)
reg     egpng   *ngp;
{
  int     allceased;

  if (ngp->ng_state == NEIGHBOR)
    n_acquired--;
  ngp->ng_state = UNACQUIRED;
  ngp->ng_status =0;
  ngp->ng_retry = 0;
  /*
   * if max # acq., set acq. interval to short
   * value for faster retry if current neigh goes down
   */
  ngp->ng_hint = (n_acquired < maxacq) ? LONGACQINT : MINHELLOINT;
  ngp->ng_htime = gatedtime + ngp->ng_hint;
  if (ngp->ng_flags & NG_BAD)
    ngp->ng_htime = gatedtime + ACQDELAY;
  egpstime();

  /*
   * If terminate signal received and all neighbors ceased then exit
   */
  if (terminate) {
    allceased = TRUE;
    for (ngp = egpngh; ngp != NULL; ngp = ngp->ng_next)
      if (ngp->ng_state == ACQUIRE_SENT || ngp->ng_state == NEIGHBOR ||
          ngp->ng_state == CEASE_SENT)
         allceased = FALSE;
    if (allceased)
      quit();
  }
}

/*
 * egphello() processes received hello packet
 */

egphello(ngp, egp)
	register     egpng   *ngp;
	register     egpkt   *egp;
{
  int     error = 0;

  if (ngp->ng_state != NEIGHBOR) {
    if (ngp->ng_state == UNACQUIRED)
      egpsacq(ngp->ng_addr, ngp->ng_myaddr, NACEASE, PROTOVIOL, egprid_h);
    ngp->ng_rcmd++;
    error = 1;
  }
  else {
    if (egp->egp_status == DOWN) {
	syslog(LOG_WARNING, "EGP neighbor %s thinks we are down",
	       inet_ntoa(ngp->ng_addr));
      ngp->ng_reach |= ME_DOWN;
    } else
      ngp->ng_reach &= ~ME_DOWN;

    switch (egp->egp_code) {
      case HELLO:
        ngp->ng_rcmd++;
        ngp->ng_rid = egprid_h;
        egpshello(ngp, HEARDU, egprid_h);
        return;
      case HEARDU:
        if (egprid_h == ngp->ng_sid)
          ngp->ng_responses |= 1;
        break;
      default:
        error = 1;
        break;
    }
  }
  if (error && tracing & TR_EXT) {
    if (!(tracing & TR_PKT))
      traceegp("egphello: recv bad pkt: ", ngp->ng_addr,
                        ngp->ng_myaddr, egp, sizeof(struct egppkt));
    printf("egphello: last recvd pkt bad, recvd in state %d\n", ngp->ng_state);
  }
  return;
}

/*
 * egppoll() processes received egp NR poll packet.
 */

egppoll(ngp, egp)
	register     egpng   *ngp;
	register     struct  egppkt  *egp;
{
  short   error = NOERROR;

  ngp->ng_paddr = ((struct egppoll *)egp)->ep_net; /* note shared net */
  ngp->ng_rcmd++;
  ngp->ng_reach &= ~ME_DOWN;              /* poll => neighbor considers me up */
  ngp->ng_responses |= 1;                 /* mark him as reachable */

  if (ngp->ng_state != NEIGHBOR) {
    if (ngp->ng_state == UNACQUIRED)
      egpsacq( ngp->ng_addr, ngp->ng_myaddr, NACEASE, PROTOVIOL, egprid_h);
      error = 100;
  }
  else
    if (egp->egp_code != 0)
      error = EBADHEAD;
  else           /* compare source net with expected net */
    if (inet_wholenetof(((struct egppoll *)egp)->ep_net) !=
        inet_wholenetof(ngp->ng_addr))
      error = EBADHEAD;
  else
    if (ngp->ng_reach & NG_DOWN)
      error = EUNSPEC;
  else {
    if (gatedtime < ngp->ng_rtime) {        /* my NR may be lost*/
      if (egprid_h != ngp->ng_rid || ++ngp->ng_rnpoll > NPOLL)
        error = EXSPOLL;
    }
    else {
      ngp->ng_rnpoll = 1;     /* new poll */
    }
  }
  if (error == NOERROR) {
    ngp->ng_rid = egprid_h;
    egpsnr(ngp, 0, egprid_h);
  }
  else {
    if (error < 100) 
      egpserr(ngp->ng_addr, ngp->ng_myaddr, egp, sizeof(struct egppoll),
                 error, ngp);
    else
      if (tracing & TR_EXT) {
        if (!(tracing & TR_PKT))
          traceegp("egppoll: recv bad pkt: ", ngp->ng_addr,
                              ngp->ng_myaddr, egp, sizeof(struct egppoll));
        printf("egppoll: last recvd pkt bad, recvd in state %d\n",
                                                        ngp->ng_state);
      }
  }
  return;
}

/*
 * egpnr() processes received NR message packet.
 * It calls rt_NRupdate() to update the routing table.
 */

egpnr( ngp, egp, egplen)
        register     egpng   *ngp;
        register     egpkt   *egp;
        int     egplen;
{
  short   error = NOERROR;

  if (ngp->ng_state != NEIGHBOR) {
    if (ngp->ng_state == UNACQUIRED)
      egpsacq(ngp->ng_addr, ngp->ng_myaddr, NACEASE, PROTOVIOL, egprid_h);
    error = 100;                            
  }
  else
    if (egp->egp_code != 0)
      error = EBADHEAD;
  else           /* compare source net with expected net */
    if (inet_wholenetof(((struct egpnr *)egp)->en_net) !=
        inet_wholenetof(ngp->ng_myaddr))
      error = EBADHEAD;
  else
    if (egp->egp_status & UNSOLICITED) {
      if (egprid_h != ngp->ng_sid)    /* wrong seq. # */
        error = 100;
      else
        if (++ngp->ng_runsol > 1)  /* too many unsol. NR */
          error = EUNSPEC;
    }
  else
    if (egprid_h != ngp->ng_sid || ngp->ng_snpoll == 0)   /* NR already recvd */
      error = 100;                    /* discard */
  else {
    /*
     * reduce next poll time slightly to
     * take account of net delays on response
     * in order to preserve sending ratio to
     * hellos. Note HELLOMARGIN*ratio already
     * added to advised min. poll interval so
     * latter still met
     */
    ngp->ng_stime +=  ngp->ng_spint - ngp->ng_hint;
    ngp->ng_snpoll = 0;             /* no more repolls required */
    ngp->ng_reach &= ~ME_DOWN;      /* update => neighbor considers me up */
    ngp->ng_responses |= 1;
    if (rt_NRupdate(ngp, egp, egplen))
      error = EBADDATA;
    else
      ngp->ng_noupdate = 0;
  }
  /*
   * poll cmds will be delayed if unsolicited updates are
   * received.  Also, unsolicited updates count for reachability.
   */
  if (error == NOERROR) {
    ngp->ng_reach &= ~ME_DOWN;      /* update => neighbor considers me up */
    ngp->ng_responses |= 1;
    if (egp->egp_status & UNSOLICITED) {
      /* reduce next poll time slightly to take account
       * of net delays on response in order to preserve sending ratio to
       * hellos. Note HELLOMARGIN*ratio already added
       * to advised min. poll interval so latter still met
       *
       * if next cmd is poll, replace with hello
       */
      if (ngp->ng_htime >= ngp->ng_stime)
        ngp->ng_htime = ngp->ng_stime;
      ngp->ng_stime += ngp->ng_hint;
      ngp->ng_snpoll = 0;         /* no more repolls required */
    }
    ngp->ng_snpoll = 0;         /* no more repolls required */
  }
  if (error != NOERROR) {
    if (error < 100)
      egpserr( ngp->ng_addr, ngp->ng_myaddr, egp, egplen, error, ngp);
    else
      if (tracing & TR_EXT) {
        if (!(tracing & TR_PKT))
          traceegp("egpnr: recv bad pkt: ",
                                ngp->ng_addr, ngp->ng_myaddr, egp, egplen);
        printf("egpnr: last recvd pkt bad, recvd in state %d\n", ngp->ng_state);
      }
  }
  return;
}

/*
 * egpsacq() sends an acquisition or cease packet.
 */

egpsacq( dst, src, code, status, id)
        struct in_addr  dst, src;    /* destination and source inet address */
        u_char  code, status;
        u_short id;
{
  struct egpacq   acqpkt;
  int length;

  acqpkt.ea_pkt.egp_ver = EGPVER;
  acqpkt.ea_pkt.egp_type = EGPACQ;
  acqpkt.ea_pkt.egp_code = code;
  acqpkt.ea_pkt.egp_status = status;
  acqpkt.ea_pkt.egp_system = htons(mysystem);
  acqpkt.ea_pkt.egp_id = htons(id);

  acqpkt.ea_hint = ntohs(MINHELLOINT);
  acqpkt.ea_pint = ntohs(MINPOLLINT);
  if (code == NAREQ || code == NACONF)
    length = sizeof(acqpkt);
  else                                    /* omit hello & poll int */
    length = sizeof(acqpkt.ea_pkt);

  acqpkt.ea_pkt.egp_chksum = 0;
  acqpkt.ea_pkt.egp_chksum = egp_cksum((char *)&acqpkt, length);

  egp_send(dst, src, (struct egppkt *)&acqpkt, length);
  return;
}

/*
 * egphello() sends a hello or I-H-U packet.
 */

egpshello( ngp, code, id)
        egpng   *ngp;
        u_char  code;
        u_short id;
{
  struct egppkt hellopkt;

  hellopkt.egp_ver = EGPVER;
  hellopkt.egp_type = EGPHELLO;
  hellopkt.egp_code = code;
  hellopkt.egp_status = (ngp->ng_reach & NG_DOWN) ? DOWN : UP;
  hellopkt.egp_system = htons(mysystem);
  hellopkt.egp_id = htons(id);
  hellopkt.egp_chksum = 0;
  hellopkt.egp_chksum = egp_cksum((char *)&hellopkt, sizeof(hellopkt));

  egp_send(ngp->ng_addr, ngp->ng_myaddr, (struct egppkt *)&hellopkt,
                 sizeof(hellopkt));
  return;
}

/*
 * egpspoll() sends an NR poll packet.
 */

egpspoll(ngp)
	register     egpng   *ngp;
{
  struct egppoll  pollpkt;
  struct in_addr  egpspolladdr;

  pollpkt.ep_pkt.egp_ver = EGPVER;
  pollpkt.ep_pkt.egp_type = EGPPOLL;
  pollpkt.ep_pkt.egp_code = 0;
  pollpkt.ep_pkt.egp_status = (ngp->ng_reach & NG_DOWN) ? DOWN : UP;
  pollpkt.ep_pkt.egp_system = htons(mysystem);
  pollpkt.ep_pkt.egp_id = htons(ngp->ng_sid);
  pollpkt.ep_unused = 0;
  egpspolladdr = gd_inet_makeaddr((int)inet_wholenetof(ngp->ng_myaddr),0,FALSE);
  pollpkt.ep_net.s_addr = egpspolladdr.s_addr;

  pollpkt.ep_pkt.egp_chksum = 0;
  pollpkt.ep_pkt.egp_chksum = egp_cksum((char *)&pollpkt,sizeof(pollpkt));

  egp_send(ngp->ng_addr, ngp->ng_myaddr, (struct egppkt *)&pollpkt,
                 sizeof(pollpkt));
  return;
}

/*
 * egpserr() sends an error packet.
 */

egpserr(dst, src, egp, length, error, ngp)
        struct in_addr  dst, src;    /* destination and source inet address */
        egpkt   *egp;   /* erroneous egp packet */
        int     length; /* length erroneous packet */
        short   error;  
        egpng   *ngp;   /* ponter to legit. neighbor table, else zero */
{
  struct egperr errpkt;

  errpkt.ee_pkt.egp_ver = EGPVER;
  errpkt.ee_pkt.egp_type = EGPERR;
  errpkt.ee_pkt.egp_code = 0;
  if (ngp && ngp->ng_state == NEIGHBOR)
    errpkt.ee_pkt.egp_status = (ngp->ng_reach & NG_DOWN) ? DOWN : UP;
  else
    errpkt.ee_pkt.egp_status = 0;
  errpkt.ee_pkt.egp_system = htons(mysystem);
  errpkt.ee_pkt.egp_id = htons(egprid_h);         /* recvd seq.# */
  errpkt.ee_rsn = htons((u_short)error);
  /*
   * copy header of erroneous egp packet
   */
  bzero((char *)errpkt.ee_egphd, sizeof(errpkt.ee_egphd));
  if (length > sizeof(errpkt.ee_egphd))
    length = sizeof(errpkt.ee_egphd);
  bcopy((char *)egp, (char *)errpkt.ee_egphd, length);

  errpkt.ee_pkt.egp_chksum = 0;
  errpkt.ee_pkt.egp_chksum = egp_cksum((char *)&errpkt, sizeof(errpkt));

  if (tracing & TR_EXT) {
    if (tracing & TR_PKT)
      printf("egpserr: send error pkt:\n");
    else
      traceegp("egpserr: send error pkt ", src, dst,
                            (struct egppkt *)&errpkt, sizeof(errpkt));
  }
  egp_send( dst, src, (struct egppkt *)&errpkt, sizeof(errpkt));
  return;
}

/*
 * egpsnr() sends an NR message packet.
 *
 * It fills in the header information, calls rt_ifcheck() to update the
 * interface status information and rt_NRnets() to fill in the reachable
 * networks.
 */

extern  u_short mysystem;
extern  int     n_interfaces;
extern  int     n_remote_nets;

egpsnr(ngp, unsol, id)
        egpng   *ngp;
        int  unsol;                     /* TRUE => set unsolicited bit */
        u_short id;
{
  int     maxsize, length;
  register  struct  egpnr   *nrp;
  struct  in_addr   egpsnraddr;

  /*
   * allocate message buffer
   */
  maxsize = sizeof(struct egpnr) + NRMAXNETUNIT* (n_interfaces + n_remote_nets);
  if (maxsize > EGPMAXPACKETSIZE - sizeof(struct ip)) {
    syslog(LOG_WARNING, "egpsnr: NR message not sent, larger than IP\n");
     return;
  }
  nrp = (struct egpnr *)malloc((unsigned)maxsize);
  if (nrp == NULL) {
    syslog(LOG_WARNING, "egpsnr:malloc: out of memory\n");
    return;
  }
  nrp->en_pkt.egp_ver = EGPVER; /* prepare static part of NR message header */
  nrp->en_pkt.egp_type = EGPNR;
  nrp->en_pkt.egp_code = 0;
  nrp->en_pkt.egp_status = (ngp->ng_reach & NG_DOWN) ? DOWN : UP;
  if (unsol)
    nrp->en_pkt.egp_status |= UNSOLICITED;
  nrp->en_pkt.egp_system = htons(mysystem);
  nrp->en_pkt.egp_id = htons(id);
  nrp->en_egw = 0;                /* no exterior gateways */
  /*
   * copy shared net address
   */
  egpsnraddr = gd_inet_makeaddr((int)inet_wholenetof(ngp->ng_myaddr), 0, FALSE);
  nrp->en_net.s_addr = egpsnraddr.s_addr;
  length = rt_NRnets(nrp, ngp->ng_myaddr); 

  if (length != ERROR) {
    nrp->en_pkt.egp_chksum = 0;
    nrp->en_pkt.egp_chksum = egp_cksum((char *)nrp, length);
    egp_send( ngp->ng_addr, ngp->ng_myaddr, (struct egppkt *)nrp, length);
  }
  else
    syslog(LOG_WARNING, "egpsnr: NR message not sent\n");
  free((char *)nrp);
  return;
}

/*
 * egp_cksum() computes the egp checksum.
 * returns checksum, note that C-GW routine
 * just returned ones complement sum
 */

egp_cksum(egp, len)
        char    *egp;           /* pointer to start egp packet */
        int     len;            /* length of egp packet */
{
  register    u_short * wd, *end_wd;
  u_long l;

  if (len & 01) {             /* pad if odd number octets */
    *(egp + len) = 0;
    len++;
  }
  wd = (u_short *)egp;       /* start word */
  end_wd = wd + (len >> 1);   /* end word */
  l = 0;
  while (wd < end_wd)
    l += *wd++;
  /*
   * convert to 16-bit ones complement sum - add 16-bit carry
   */
  l = (l & 0xffff) + ((l & 0xffff0000) >> 16);
  l = (l & 0xffff) + ((l & 0xffff0000) >> 16);
  return (((~l) & 0xffff));
}

/*
 * egp_send() sends an egp packet.
 */

egp_send(dst, src, egp, length)
        struct in_addr  dst, src;   /* destination and source inet address */
        struct egppkt   *egp;   /* pointer to start of egp packet */
        int     length;         /* length in octets of egp packet */

{
  struct sockaddr_in egpsaddr;
  struct interface   *ifp;
  int     error = FALSE;

  bzero((char *)&egpsaddr, sizeof (egpsaddr));
  egpsaddr.sin_family = AF_INET;
  /*
   * find interface to send packet on
   */
  egpsaddr.sin_addr = src;
  ifp = if_withnet(&egpsaddr);
  if (ifp == NULL) {
    syslog(LOG_ERR, "egp_send: no interface for source address %s\n",
                          inet_ntoa(src));
     error = TRUE;
  }
  else {
    extern int egp_socket;
    egpsaddr.sin_addr = dst;
    if (sendto(egp_socket, (char *)egp, length, 0,
               &egpsaddr, sizeof(egpsaddr)) < 0) {
      if (tracing & TR_EXT || tracing & TR_PKT) {
        TRACE_EXT("egp_send: sendto: %s %d bytes\n", inet_ntoa(dst), length);
        perror("egp_send: sendto:");
        error = TRUE;
      }
    syslog(LOG_ERR, "egp_send: sendto() error sending %d bytes to %s: %m",
           length, inet_ntoa(dst));
    }
  }
  if (!error) {
    TRACE_EGPPKT(EGP SENT, src, dst, egp, length);
  } else {
    TRACE_EGPPKT(EGP *NOT* SENT, src, dst, egp, length);
  }

  return;
}

/*
 * egpcease() initiates the sending of an egp neighbor cease.
 */

egpcease( ngp, rsn)
	register     egpng   *ngp;
        u_char  rsn;
{
  egpsacq( ngp->ng_addr, ngp->ng_myaddr, NACEASE, rsn, ngp->ng_sid);
  if (ngp->ng_state == NEIGHBOR) {
    n_acquired--;
    TRACE_EXT("egpcease: cease to %s, reason %d, %s\n",
                                inet_ntoa(ngp->ng_addr), rsn, strtime);
  }
  ngp->ng_state = CEASE_SENT;
  ngp->ng_retry = 0;
  if (ngp->ng_state != NEIGHBOR)
    ngp->ng_hint = MINHELLOINT;
  ngp->ng_htime = gatedtime + ngp->ng_hint;
  ngp->ng_status = rsn;           /* save reason for retransmission */
  return;
}

/*
 * egpallcease() sends Ceases to all neighbors when going down (when SIGTERM
 * received).
 *
 * Global variables:
 * terminate - set here, tested by egpstunacq()
 */

egpallcease()
{
  register egpng *ngp;
  int     ceasesent = FALSE;

  getod();                        /* time of current interrupt */
  /*
   * scan neighbor state tables
   */
  for (ngp = egpngh; ngp != NULL; ngp = ngp->ng_next) {
    if ((ngp->ng_state == NEIGHBOR) || (ngp->ng_state == ACQUIRE_SENT)) {
      egpcease(ngp, GODOWN);
      ceasesent = TRUE;
    }
  }
  if (ceasesent)
    terminate = TRUE;
  else
    quit();
}
