
/*
 *   CENTER FOR THEORY AND SIMULATION IN SCIENCE AND ENGINEERING
 *                      CORNELL UNIVERSITY
 *
 *      Portions of this software may fall under the following
 *      copyrights: 
 *
 *      Copyright (c) 1983 Regents of the University of California.
 *      All rights reserved.  The Berkeley software License Agreement
 *      specifies the terms and conditions for redistribution.
 *
 *      The Copyright (c) of the EGPUP User Process, Paul Kirton and
 *      ISI, August 23, 1984
 *
 */

/*
 *      The following acknowledgments and thanks apply:
 *
 *      Mike Petry (petry@trantor.umd.edu), University of Maryland, for
 *      development of the HELLO routines used in this program.
 *
 *      Craig Partridge (craig@loki.bbn.com), BBN Laboratories, for
 *      thorough linting, profiling, and performance enhancements to
 *      this program.
 *
 *      Charles Hedrick (hedrick@topaz.rutgers.edu),
 *      Chris Maio (chris@columbia.edu), Sergio Heker (heker@jvnca.csc.org),
 *      and any other BETA testers who have helped in finding bugs.
 *
 *	Mark Fedor (now fedor@nisc.nyser.net) for the development and
 *	maintenance up to release 1.3.1.  Also for his continuing
 *	advice.
 *
 *	Thomas Narten <narten@purdue.edu> for his analysis and
 *	modifications to the EGP and other helpful improvements.
 *
 *
 *
 *  GATED - based on Kirton's EGP, UC Berkeley's routing daemon (routed),
 *          and DCN's HELLO routing Protocol.  HELLO output routines
 *          were developed and originally written at the University of
 *          Maryland by Mike Petry.
 * 
 *  December 5, 1986 - Mark Fedor
 *
 * $Header: rt_egp.c,v 1.3.1.17 88/01/09 19:00:40 jch Locked $
 *
 */

#ifndef	lint
static char *rcsid = "$Header: rt_egp.c,v 1.3.1.17 88/01/09 19:00:40 jch Locked $";
#endif	not lint

/* 
 * rt_egp.c
 *
 * EGP route update and processing and preparation functions.
 *
 * Functions: rt_NRnets, rt_NRupdate
 */

#include "include.h"

/*
 * rt_NRnets() prepares the network part of the EGP Network Reachability 
 * update message with respect to the shared network of the EGP peer.
 * This only includes the networks in the interior routing table (direct 
 * networks, and remote networks of non-routing gateways of this autonomous
 * system) other than the net shared by the EGP peer. If the user has
 * specified that only certain networks are allowed to be advised all others 
 * are excluded from outgoing NR update messages.
 * If the interior routing table includes other interior gateways on the
 * network shared with the EGP peer (i.e. indirect neighbors), they are
 * included in updates as the appropriate first hop to their attached
 * networks.
 * This function checks the status of routes and if down sets the
 * distance as unreachable.
 *
 * Returns the length of the EGP NR packet in octets or ERROR if an error
 * has occurred.
 */

rt_NRnets(nrpkt, sharednet_addr)
        struct  egpnr  *nrpkt;          /* start of NR message */
        struct in_addr  sharednet_addr; /* internet address of interface to
                                           shared net of EGP neighbor */
{
  struct  rt_entry  *rt;
  struct  rthash *rh;
  struct  net_order {       /* temporary linked list for ordering nets */
         struct net_order *next;
         struct in_addr  net;            /* net # */
         struct in_addr  gateway;
         int     distance;
  } *start_net, *free_net;

  register  struct  net_order *net_pt, *this_net;  /* current search point */
  int       shared_net, n_bytes;
  struct in_addr current_gw;
  register  u_char  *nrp;                   /* next octet of NR message */
  u_char    *n_distance, *distance, *n_nets;
  int       this_metric;
  struct interface *tifp;
  struct sockaddr_in tsock;

  shared_net = inet_wholenetof(sharednet_addr);

  bzero((char *)&tsock, sizeof(tsock));
  tsock.sin_family = AF_INET;
  tsock.sin_addr.s_addr = sharednet_addr.s_addr;
  if ((tifp = if_ifwithaddr(&tsock)) <= (struct interface *)0) {
    printf("No interface for egp, %s wanted\n", inet_ntoa(sharednet_addr));
    return(ERROR);
  }

  /*
   * Reorder the interior routes as required for the NR message with respect to
   * the given shared net. Uses a temporary linked list terminated by NULL
   * pointer. The first element of the list is a dummy so insertions can be done
   * before the first true entry. The route status is checked and if down the
   * distance is set as unreachable.  The required order groups nets by gateway
   * and in order of increasing metric. This gateway is listed first (with all
   * nets not reached by gateways on the shared net) and then neighbor gateways
   * on the shared net, in any order. As there are few nets to be reported by a
   * stub gateway, each route is copied from the interior routing table and
   * inserted in the temporary reordered list using a linear search.
   */
  start_net = (struct net_order *)malloc((unsigned)((n_remote_nets +
                                    n_interfaces) * sizeof(struct net_order)));
  if (start_net == NULL) {
    syslog(LOG_ERR, "rt_NRnets: malloc: out of memory\n");
    return(ERROR);
  }
  start_net->next = NULL;
  /*
   * ensures first gateway listed is self
   */
  start_net->gateway = sharednet_addr;
  free_net = start_net + 1;   /* first element dummy to ease insertion code */

  /*
   * check all interior routes of route table
   */
  for (rh = nethash; rh < &nethash[ROUTEHASHSIZ]; rh++)
    for (rt = rh->rt_forw; rt != (struct rt_entry *)rh; rt = rt->rt_forw) {
      if ((rt->rt_state & RTS_INTERIOR) == 0)
        continue;
      /*
       *  extra sanity check cause we don't want to EGP out
       *  subnets.
       */
      if (htonl(inet_wholenetof(sock_inaddr(&rt->rt_dst))) !=
          sock_inaddr(&rt->rt_dst).s_addr)
        continue;
      TRACE_UPD("net %s - ",inet_ntoa(sock_inaddr(&rt->rt_dst)));
      if (sock_inaddr(&rt->rt_dst).s_addr == DEFAULTNET) {
        TRACE_UPD("not sending default.\n");
        continue;
      }
      /*
       * ignore shared net
       */
      if (inet_wholenetof(sock_inaddr(&rt->rt_dst)) == shared_net) {
        TRACE_UPD("shared net.\n");
        continue;
      }
      /*
       * ignore nets not to be advised
       */
      if (adlist) {     /* list of nets to be advertized */
        register struct advlist *ad;
        register int OK = 0;

        for (ad = adlist; ad; ad = ad->next)
          if (sock_inaddr(&rt->rt_dst).s_addr == ad->destnet.s_addr)
            OK++;
        if (!OK) {
          TRACE_UPD("not advise\n");
          continue;
        }
      }
      /*
       * check to see if we already have this net here.
       * we may have multiple interfaces to subnets of it.
       */
      for (net_pt = start_net->next; net_pt ; net_pt = net_pt->next) {
        register u_long net = inet_wholenetof(sock_inaddr(&rt->rt_dst));

        if ((net) == inet_wholenetof(net_pt->net)) {
          /*
           * already advertising this net
           * but make sure we advertise shortest metric
           */
          if ((rt->rt_metric/ETOTD_CONV) < net_pt->distance)
            net_pt->distance = rt->rt_metric / ETOTD_CONV;
          TRACE_UPD("already in list\n");
          goto Continue;
        }
      }
      /*
       * routes flagged as EGP or Interface (== RTPROTO_DIRECT) are not
       * restricted and we use the true metric.  If it is another
       * type of route, we see if it is restricted and what metric
       * to use.
       */
      if ((rt->rt_proto != RTPROTO_EGP) && (rt->rt_proto != RTPROTO_DIRECT)) {
        if (rt->rt_proto & (RTPROTO_RIP|RTPROTO_HELLO)) {
          if ((tifp->int_flags & IFF_ANNOUNCE) ||
              (glob_announcethesenets != 0)) {
            if (rt->rt_announcelist == (struct restrictlist *)NULL) {
              TRACE_UPD(" restricted\n");
              goto Continue;
            }
            if ((rt->rt_announcelist->rproto & RTPROTO_EGP) == 0) {
              TRACE_UPD(" restricted\n");
              goto Continue;
            }
            if ((rt->rt_proto & RTPROTO_RIP) &&
                (rt->rt_metric >= DELAY_INFINITY))
              this_metric = HOPCNT_INFINITY;
            else
              if ((rt->rt_proto & RTPROTO_HELLO) &&
                  (rt->rt_metric >= DELAY_INFINITY))
                this_metric = HOPCNT_INFINITY;
            else
              this_metric = rt->rt_announcelist->regpmetric;
          }
          else
            if ((tifp->int_flags & IFF_NOANNOUNCE) ||
                (glob_donotannounce != 0)) {
              if ((rt->rt_noannouncelist != (struct restrictlist *)NULL) &&
                  (rt->rt_noannouncelist->rproto & RTPROTO_EGP)) {
                TRACE_UPD(" restricted\n");
                goto Continue;
              }
              else {
                if (rt->rt_noannouncelist != (struct restrictlist *)NULL)
                  this_metric = conv_factor;
                else {
                  TRACE_UPD(" restricted\n");
                  goto Continue;
                }
              }
            }
            else {
              /*
               * we have no announce or noannounce restrictions.
               * if our soft restriction egpnetsreachable comes
               *  through we can advertise it.  Else Nix the route.
               */
              if (adlist)
                this_metric = conv_factor;
              else {
                TRACE_UPD("restricted\n");
                goto Continue;
              }
            }
          }
          else {
            TRACE_UPD("restricted\n");
            goto Continue;
          }
        }
        else
          if (rt->rt_flags & RTF_UP)
            this_metric = rt->rt_metric / ETOTD_CONV;
          else
            this_metric = HOPCNT_INFINITY;
        /*
         * committed to advertising net
         */
        this_net = free_net++;
        this_net->net.s_addr = sock_inaddr(&rt->rt_dst).s_addr;
        this_net->distance = this_metric;
        /*
         * assign gw on shared net
         */
        if (inet_wholenetof(sock_inaddr(&rt->rt_router)) != shared_net)
          this_net->gateway = sharednet_addr;     /* gw is self */
        else                                      /* gw is neighbor */
          this_net->gateway = sock_inaddr(&rt->rt_router);

        TRACE_UPD(" metric %d ..", this_net->distance);
        /*
         * insert net in ordered list
         */
        for (net_pt = start_net; ; net_pt = net_pt->next) {
          if (net_pt->next == NULL)
            break;        /* end list */
          if (this_net->gateway.s_addr == net_pt->next->gateway.s_addr) {
            if (this_net->distance <= net_pt->next->distance)
              break;
          }
          else
            if (this_net->gateway.s_addr == net_pt->gateway.s_addr)
              break;
        }
        /*
         * insert this net after search net
         */
        this_net->next = net_pt->next;
        net_pt->next = this_net;
        TRACE_UPD("added to update dist. %d\n", this_net->distance);
Continue: ;
    }                               /* end for all interior routes */

  /*
   * copy nets into NR message
   */
  nrpkt->en_igw = 0;              /* init # interior gateways */
  nrp = (u_char *)(nrpkt + 1);    /* start nets part NR msg */
  current_gw.s_addr = 0;          /* ensure first gateway addr copied */
  for (net_pt = start_net->next; net_pt != NULL; net_pt = net_pt->next) {
    if (net_pt->gateway.s_addr != current_gw.s_addr) { /* new gateway */
      current_gw.s_addr = net_pt->gateway.s_addr;
#if BSD42 || SUN3_3PLUS
      if (in_isa(current_gw.s_net))
#else
      if (in_isa(current_gw.s_addr))
#endif
        n_bytes = 3;
      else 
#if BSD42 || SUN3_3PLUS
        if (in_isb(current_gw.s_net))
#else
        if (in_isb(current_gw.s_addr))
#endif
          n_bytes = 2;
        else
          n_bytes = 1;
      bcopy((char *)&current_gw + 4 - n_bytes, (char *)nrp, n_bytes);
      nrp += n_bytes;
      nrpkt->en_igw++;
      n_distance = nrp++;
      *n_distance = 1;
      distance = nrp++;
      *distance = net_pt->distance;
      n_nets = nrp++;
      *n_nets = 1;
    }
    else
      if (net_pt->distance != *distance) {   /* new distance */
        (*n_distance)++;
        distance = nrp++;
        *distance = net_pt->distance;
        n_nets = nrp++;
        *n_nets = 1;
      }
      else
        (*n_nets)++;

#if BSD42 || SUN3_3PLUS
    if (in_isa(net_pt->net.s_net))
#else
    if (in_isa(net_pt->net.s_addr))
#endif
      n_bytes = 1;
    else
#if BSD42 || SUN3_3PLUS
      if (in_isb(net_pt->net.s_net))
#else
      if (in_isb(net_pt->net.s_addr))
#endif
        n_bytes = 2;
      else
        n_bytes = 3;
    bcopy((char *)&net_pt->net.s_addr, (char *)nrp, n_bytes);
    nrp += n_bytes;
  }                                      /* end for each net */
  free((char *)start_net);
  return(nrp - (u_char *)nrpkt);         /* length of NR message */
}

/*
 * rt_NRupdate() updates the exterior routing tables on receipt of an NR
 * update message from an EGP neighbor. It first checks for valid NR counts 
 * before updating the routing tables.
 *
 * EGP Updates are used to update the exterior routing table if one of the
 * following is satisfied:
 *   - No routing table entry exists for the destination network and the
 *     metric indicates the route is reachable (< 255).
 *   - The advised gateway is the same as the current route.
 *   - The advised distance metric is less than the current metric.
 *   - The current route is older (plus a margin) than the maximum poll
 *     interval for all acquired EGP neighbors. That is, the route was
 *     omitted from the last Update.
 *
 * Returns 1 if there is an error in NR message data, 0 otherwise.
 */

#define CLAA 1
#define CLAB 2
#define CLAC 3

rt_NRupdate(ngp, pkt, egplen)
        struct egpngh  *ngp;            /* pointer to neighbor state table */
        struct egppkt  *pkt;
        int             egplen;         /* length EGP NR packet */
{
  register  u_char  *nrb;
  struct egpnr *nrp = (struct egpnr *)pkt;
  struct sockaddr_in      destination, gateway;
  u_char  gw[4];                          /* gateway internet address */
  int     gw_class,
          net_class,
          ng,
          nd,
          nn,
          n_gw,
          n_dist,
          n_net,
          dist,
          state,
          checkingNR = TRUE,
          change = FALSE;
  struct rt_entry *rt;

  bzero((char *)&destination, sizeof(destination));
  bzero((char *)&gateway, sizeof(gateway));
  gateway.sin_family = AF_INET;
  destination.sin_family = AF_INET;

  /*
   * check class of shared net
   */
  *(u_long *)gw = nrp->en_net.s_addr;     /* set net part of gateways */
  if (in_isa(gw[0]))
    gw_class = CLAA;
  else if (in_isb(gw[0]))
    gw_class = CLAB;
  else if (in_isc(gw[0]))
    gw_class = CLAC;
  else
    return(1);              /* NR message error */

  n_gw = nrp->en_igw + nrp->en_egw;

  /*
   * First check NR message for valid counts, then repeat and update routing
   * tables
   */
repeat:
  nrb = (u_char *)nrp + sizeof(struct egpnr);  /* start first gw */

  for (ng = 0; ng < n_gw; ng++) {         /* all gateways */
    switch (gw_class) {         /* fill gateway local address */
      case CLAA:      gw[1] = *nrb++;
      case CLAB:      gw[2] = *nrb++;
      case CLAC:      gw[3] = *nrb++;
    }
    gateway.sin_addr.s_addr = *(u_long *)gw;
    n_dist = *nrb++;

    for (nd = 0; nd < n_dist; nd++) {   /* all distances this gateway */
      dist = (*nrb++) * ETOTD_CONV;
      n_net = *nrb++;
      for (nn = 0; nn < n_net; nn++) {  /* all nets this distance */
        if (in_isa(*nrb))
          net_class = CLAA;
        else if (in_isb(*nrb))
          net_class = CLAB;
        else if (in_isc(*nrb))
          net_class = CLAC;
        else 
          return(1);
        destination.sin_addr.s_addr = 0;    /* zero unused bytes*/
        bcopy((char *)nrb, (char *)&destination.sin_addr.s_addr, net_class);
        nrb += net_class;
        if (checkingNR) {           /* first check counts only */
          if (nrb > (u_char *)nrp + egplen + 1)
            return(1);          /* erroneous counts in NR */
        }
        else {                      /* update routing table */
          if (gateway.sin_addr.s_addr == ngp->ng_myaddr.s_addr)
            continue;
          /*
           * check for internal route
           */
          state = 0;
          if ((rt = rt_lookup((int)INTERIOR, &destination)) != NULL) {
            /*
             * assume INTERIOR route is better ,
             * if it's there at all.
             * 
             *  Extra sanity check to keep duplicate routes
             *  out of the kernel.   Makes things neater.
             */
            while (rt = rt_lookup((int)EXTERIOR, &destination))
              rt_delete(rt, KERNEL_INTR);
            continue;
#ifdef notdef                       
            if (gateway.sin_addr.s_addr == sock_inaddr( &rt->rt_router))
              continue;       /*ignore same route */
            else
              state = RTS_NOTINSTALL;
#endif                         
          }
          rt = rt_lookup((int)EXTERIOR, &destination);
          if (rt == NULL) {       /* new route */
            if (dist >= TDHOPCNT_INFINITY)
              continue;
            (void) rt_add((int)EXTERIOR, &destination, &gateway, dist, state, 
                   RTPROTO_EGP, RTPROTO_EGP);
            change = TRUE;
          }
          else {                  /* existing route */
            if (equal(&rt->rt_router, &gateway)) {     /* same gw */
              if (dist < TDHOPCNT_INFINITY) {
                if (rt_change(rt, &gateway, dist, RTPROTO_EGP, RTPROTO_EGP))
                  change = TRUE;
              }
              else {
                rt->rt_metric = TDHOPCNT_INFINITY;
                rt->rt_timer = rt_maxage - (FINAL_DELETE-EXPIRE_TIME);
              }
            }
            else {              /* different gateway */
              if ((dist < rt->rt_metric) || (dist >= rt->rt_metric &&
                  rt->rt_timer > maxpollint && !(rt->rt_state & RTS_CHANGED))) {
                (void) rt_change(rt, &gateway, dist, RTPROTO_EGP, RTPROTO_EGP);
                change = TRUE;
              }
            }
          }       /* end else existing route */
        }           /* end else update routing table */
      }               /* end for all nets */
    }                   /* end for all distances */
  }                       /* end for all gateways */
  if (checkingNR) {
    if (nrb > (u_char *)nrp + egplen)
      return(1);                      /* erroneous counts */
    else
      checkingNR = FALSE;
    goto repeat;
  }
  if (rt_default_status == INSTALLED) {
    rt_default("DELETE");
    change = TRUE;
  }
  if (change && tracing & TR_RT)
    printf("rt_NRupdate: above routes from %s updated %s\n",
                        inet_ntoa(ngp->ng_addr), strtime);
  return(0);
}
